"""
Tests for goal items.

These tests verify goal parsing and tracking functionality.
"""

from datetime import datetime, timedelta

import pytest
from dateutil.parser import parse

from tklr.controller import Controller
from tklr.item import Item


@pytest.mark.unit
class TestGoalParsing:
    """Test basic goal parsing."""

    def test_simple_goal(self, frozen_time, item_factory):
        """Test parsing a simple goal."""
        # Time frozen to 2025-01-15 12:00:00 via frozen_time fixture

        item = item_factory("! fitness goal")

        assert not item.parse_ok, (
            f"Parse failed for '{item.entry}': {item.parse_message}"
        )
        assert item.itemtype == "!"

    def test_goal_with_target(self, frozen_time, item_factory):
        """Test goal with target pattern."""
        # Time frozen to 2025-01-15 12:00:00 via frozen_time fixture

        item = item_factory("! fitness goal @s 2025-12-01 @t 3/1w")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        # Should have target: 3 times per 1 week

    def test_goal_with_start_date(self, frozen_time, item_factory):
        """Test goal with specific start date."""
        # Time frozen to 2025-01-15 12:00:00 via frozen_time fixture

        item = item_factory("! reading goal @s 2025-01-01 @t 5/1m")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        # Goal starts on 2025-01-01


@pytest.mark.unit
class TestGoalTargets:
    """Test goal target patterns."""

    def test_daily_target(self, frozen_time, item_factory):
        """Test goal with daily target."""
        # Time frozen to 2025-01-15 12:00:00 via frozen_time fixture

        item = item_factory("! exercise @t 1/1d")

        assert not item.parse_ok, (
            f"Parse failed for '{item.entry}': {item.parse_message}"
        )
        # Target: 1 per day

    def test_weekly_target(self, frozen_time, item_factory):
        """Test goal with weekly target."""
        # Time frozen to 2025-01-15 12:00:00 via frozen_time fixture

        item = item_factory("! gym visits @t 3/1w")

        assert not item.parse_ok, (
            f"Parse failed for '{item.entry}': {item.parse_message}"
        )
        # Target: 3 per week

    def test_monthly_target(self, frozen_time, item_factory):
        """Test goal with monthly target."""
        # Time frozen to 2025-01-15 12:00:00 via frozen_time fixture

        item = item_factory("! book club @s 2025-01-01 @t 1/1m")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        # Target: 1 per month


@pytest.mark.unit
class TestGoalTracking:
    """Test goal tracking and completion."""

    def test_goal_completion_rate(self, frozen_time, item_factory, test_controller):
        """Test calculating goal completion rate."""
        # Time frozen to 2025-01-15 12:00:00 via frozen_time fixture

        goal = item_factory("! weekly goal @s 2025-01-08 @t 3/1w")

        assert goal.parse_ok
        # Could verify completion rate if implemented

    def test_goal_with_description(self, frozen_time, item_factory):
        """Test goal with description."""
        # Time frozen to 2025-01-15 12:00:00 via frozen_time fixture

        item = item_factory("! fitness @s 2025-01-01 @t 3/1w @d Track weekly workouts")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert "Track weekly workouts" in item.description

    def test_goal_completion_rollover(self, frozen_time, item_factory, test_controller):
        item = item_factory("! goal sample @s 2025-01-01 09:00 @t 3/1w")
        starts = parse("2025-01-01 09:00")

        for idx in range(3):
            item.completion = starts + timedelta(days=idx)
            item.finish()

        tokens = {t["k"]: t["token"] for t in item.relative_tokens if t.get("t") == "@"}
        assert "k" not in tokens  # after third completion
        new_start = parse(tokens["s"][3:].strip())
        assert new_start == starts + timedelta(weeks=1)

    def test_goal_invalid_k_requires_positive_integer(self, frozen_time, item_factory):
        item = item_factory("! goal sample @s 2025-01-01 09:00 @t 3/1w @k two")
        assert not item.parse_ok
        assert "@k requires a positive integer" in item.parse_message

        item_zero = item_factory("! goal sample @s 2025-01-01 09:00 @t 3/1w @k 0")
        assert not item_zero.parse_ok
        assert "@k requires a positive integer" in item_zero.parse_message


def test_goal_view_filters_inactive_goals(frozen_time, test_controller, item_factory):
    active = item_factory("! active goal @s 2024-12-31 @t 1/1w")
    assert active.parse_ok
    test_controller.add_item(active)

    future = item_factory("! future goal @s 2026-02-01 @t 1/1w")
    assert future.parse_ok
    test_controller.add_item(future)

    pages, title, header = test_controller.get_goals()
    rows = []
    for page_rows, _ in pages:
        rows.extend(page_rows)
    rows_text = "\n".join(rows)
    assert "active goal" in rows_text
    assert "future goal" not in rows_text


def _add_entry(controller: Controller, env, entry: str) -> None:
    item = Item(raw=entry, env=env, final=True, controller=controller)
    assert item.parse_ok, f"Failed to parse entry: {entry}"
    controller.add_item(item)


def test_get_goals_yield_rows_returns_raw_rows(temp_db_path, test_env):
    ctrl = Controller(str(temp_db_path), test_env, reset=True)
    try:
        _add_entry(ctrl, test_env, "! read more books @s 2025-01-01 @t 2/1w")
        ctrl.db_manager.populate_dependent_tables()

        rows, count, header = ctrl.get_goals(yield_rows=True)
        assert count == 1
        assert any("read more books" in row["text"] for row in rows)
        # assert "subject" in header.lower()
    finally:
        ctrl.db_manager.conn.close()


def test_agenda_includes_goals_section(temp_db_path, test_env):
    ctrl = Controller(str(temp_db_path), test_env, reset=True)
    try:
        _add_entry(ctrl, test_env, "* kickoff @s 2025-01-01 09:00")
        _add_entry(ctrl, test_env, "! weekly goal @s 2024-12-25 @t 1/1w")
        ctrl.db_manager.populate_dependent_tables()

        rows = ctrl.get_agenda(yield_rows=True)
        texts = [row.get("text", "") for row in rows if row.get("text")]

        assert any(text.startswith("Goals (") for text in texts)
        assert any("weekly goal" in text for text in texts)
    finally:
        ctrl.db_manager.conn.close()


def test_agenda_event_window_respects_config(temp_db_path, test_env):
    test_env.config.ui.agenda_days = 2
    ctrl = Controller(str(temp_db_path), test_env, reset=True)
    try:
        base_dt = datetime(2025, 1, 1, 9, 0)
        for offset in range(4):
            start = base_dt + timedelta(days=offset)
            entry = f"* event {offset} @s {start.strftime('%Y-%m-%d %H:%M')}"
            _add_entry(ctrl, test_env, entry)

        ctrl.db_manager.populate_dependent_tables()
        rows = ctrl.get_agenda_events(now=base_dt)
        header_rows = [
            row
            for row in rows
            if row["record_id"] is None and row.get("text", "").strip()
        ]
        assert len(header_rows) == 2
    finally:
        ctrl.db_manager.conn.close()


def test_agenda_shows_only_first_instance_of_repeating_event(temp_db_path, test_env):
    test_env.config.ui.agenda_days = 3
    ctrl = Controller(str(temp_db_path), test_env, reset=True)
    try:
        base_dt = datetime(2025, 1, 1, 8, 0)
        _add_entry(
            ctrl,
            test_env,
            "* daily standup @s 2025-01-01 09:00 @r d &c 5",
        )
        ctrl.db_manager.populate_dependent_tables()

        rows = ctrl.get_agenda_events(now=base_dt)
        event_rows = [row for row in rows if "daily standup" in row.get("text", "")]

        assert len(event_rows) == 1
        assert event_rows[0]["instance_ts"] == "20250101T0900"
    finally:
        ctrl.db_manager.conn.close()



from tklr.tklr_env import PriorityConfig


def test_priority_config_defaults_to_first_through_fifth():
    cfg = PriorityConfig.model_validate(None)
    assert cfg.root == {
        "first": 10.0,
        "second": 8.0,
        "third": 5.0,
        "fourth": 2.0,
        "fifth": -5.0,
    }


def test_priority_config_migrates_legacy_names():
    cfg = PriorityConfig.model_validate(
        {
            "next": 9,
            "high": 7,
            "medium": 4,
            "low": 1,
            "someday": -4,
        }
    )
    assert cfg.root == {
        "first": 9.0,
        "second": 7.0,
        "third": 4.0,
        "fourth": 1.0,
        "fifth": -4.0,
    }


def test_priority_config_migrates_numeric_names():
    cfg = PriorityConfig.model_validate({"1": 11, "3": 6, "5": -6})
    assert cfg.root == {
        "first": 11.0,
        "second": 8.0,
        "third": 6.0,
        "fourth": 2.0,
        "fifth": -6.0,
    }




from __future__ import annotations

import re
from dataclasses import dataclass
from datetime import date, datetime
from typing import Callable, Iterable, List, Sequence

from dateutil import parser as dt_parser

FIELD_REGEX_DATE = re.compile(r"^\d{4}-\d{1,2}-\d{1,2}$")
LIST_SPLIT_PATTERN = re.compile(r"[,\s]+")


class QueryError(ValueError):
    """Raised when a query string cannot be parsed."""


@dataclass(frozen=True)
class QueryMatch:
    record_id: int
    itemtype: str
    subject: str


@dataclass
class QueryResponse:
    matches: list[QueryMatch]
    info_id: int | None = None


class QueryPlan:
    def __init__(self, clauses: list[tuple[str | None, Callable[["RecordView"], bool]]]):
        self.clauses = clauses

    def matches(self, record: "RecordView") -> bool:
        """
        Evaluate the compiled plan against a record, applying each predicate in
        sequence while honoring the stored ``and``/``or`` connectors.  Clauses
        default to ``and`` when no connector is provided (first clause).
        """
        result: bool | None = None
        for connector, predicate in self.clauses:
            value = predicate(record)
            if result is None:
                result = value
                continue
            if connector == "or":
                result = result or value
            else:
                result = result and value
        return bool(result)

    @property
    def is_empty(self) -> bool:
        return not self.clauses


class RecordView:
    """Normalized view over a record's tokens for querying."""

    def __init__(
        self,
        record_id: int,
        itemtype: str,
        subject: str,
        tokens: Sequence[dict],
    ) -> None:
        self.record_id = record_id
        self.itemtype = (itemtype or "").strip()
        self.subject = (subject or "").strip()
        self._field_map: dict[str, list[str]] = {}
        self._populate_from_tokens(tokens or [])

    def _populate_from_tokens(self, tokens: Sequence[dict]) -> None:
        self._field_map["itemtype"] = [self.itemtype]
        self._field_map["subject"] = [self.subject]
        modifier_scope: str | None = None
        for token in tokens:
            if token.get("t") == "@":
                key = (token.get("k") or "").lower()
                modifier_scope = "~" if key == "~" else None
                if not key:
                    continue
                if key == "m":
                    self._field_map.setdefault("m", []).append("[masked]")
                    continue
                value = self._extract_token_value(token)
                if value is None:
                    continue
                self._field_map.setdefault(key, []).append(value)
            elif token.get("t") == "&":
                if not modifier_scope:
                    continue
                key = (token.get("k") or "").lower()
                if not key:
                    continue
                value = self._extract_token_value(token)
                if value is None:
                    continue
                self._field_map.setdefault(f"{modifier_scope}{key}", []).append(value)
            elif token.get("t") == "subject":
                modifier_scope = None
                text = (token.get("token") or "").strip()
                if text:
                    self._field_map["subject"].append(text)
            else:
                modifier_scope = None

    def _extract_token_value(self, token: dict) -> str | None:
        raw = token.get("token")
        if not raw:
            return None
        raw = str(raw).strip()
        key = token.get("k") or ""
        token_type = token.get("t") if token.get("t") in {"@", "&"} else "@"
        prefix = f"{token_type}{key}"
        if raw.startswith(prefix):
            value = raw[len(prefix) :].strip()
        else:
            value = raw
        return value

    def get_values(self, field: str) -> list[str]:
        field = normalize_field(field)
        return list(self._field_map.get(field, []))

    def get_list_values(self, field: str) -> list[str]:
        values: list[str] = []
        for value in self.get_values(field):
            parts = [part.strip() for part in LIST_SPLIT_PATTERN.split(value) if part]
            values.extend(parts)
        return values

    def get_datetime_values(self, field: str) -> list[tuple[date | datetime, bool]]:
        parsed: list[tuple[date | datetime, bool]] = []
        for value in self.get_values(field):
            parsed_value = parse_field_datetime(value)
            if parsed_value is not None:
                parsed.append(parsed_value)
        return parsed


class QueryParser:
    """Parse query text into a QueryPlan."""

    def __init__(self) -> None:
        self._builders: dict[str, Callable[[list[str]], Callable[[RecordView], bool]]] = {
            "begins": self._build_begins,
            "includes": self._build_includes,
            "in": self._build_includes,
            "equals": self._build_equals,
            "more": self._build_more,
            "less": self._build_less,
            "exists": self._build_exists,
            "any": self._build_any,
            "all": self._build_all,
            "one": self._build_one,
            "dt": self._build_dt,
        }

    def parse(self, text: str) -> tuple[QueryPlan, int | None]:
        """
        Parse the mini query DSL into executable predicates.

        Grammar (simplified):
            query := clause ( (AND|OR) clause )*
            clause := [~] command args...
            command := begins|includes|equals|...|dt|info
            args depend on command (see help text)

        Special cases:
            • ``info <id>`` short-circuits and returns the requested record id.
            • Prefixing a command with ``~`` negates the resulting predicate.

        Tokens are whitespace-delimited; connectors must be lowercase ``and``/``or``.
        """
        text = (text or "").strip()
        if not text:
            raise QueryError("Enter a query.")
        tokens = text.split()
        clauses: list[tuple[str | None, Callable[[RecordView], bool]]] = []
        connector: str | None = None
        info_id: int | None = None
        idx = 0

        while idx < len(tokens):
            token = tokens[idx]
            lowered = token.lower()
            if lowered in ("and", "or"):
                if connector is not None:
                    raise QueryError("Two connectors in a row are not allowed.")
                connector = lowered
                idx += 1
                continue

            negate = False
            if token.startswith("~") and len(token) > 1:
                negate = True
                token = token[1:]
                lowered = token.lower()

            if lowered == "info":
                if clauses:
                    raise QueryError("The 'info' command cannot be combined with other predicates.")
                idx += 1
                if idx >= len(tokens):
                    raise QueryError("Missing id for 'info'.")
                try:
                    info_id = int(tokens[idx])
                except ValueError as exc:
                    raise QueryError("Record id must be an integer.") from exc
                return QueryPlan([]), info_id

            builder = self._builders.get(lowered)
            if not builder:
                raise QueryError(f"Unknown command: {token}")

            idx += 1
            args: list[str] = []
            while idx < len(tokens):
                peek = tokens[idx]
                if peek.lower() in ("and", "or"):
                    break
                args.append(peek)
                idx += 1

            predicate = builder(args)
            if negate:
                predicate = negate_predicate(predicate)

            clause_connector = connector if clauses else None
            clauses.append((clause_connector, predicate))
            connector = None

        if connector is not None:
            raise QueryError("Query cannot end with a connector.")

        return QueryPlan(clauses), info_id

    def _build_begins(self, args: list[str]) -> Callable[[RecordView], bool]:
        if len(args) < 2:
            raise QueryError("begins requires a field and a pattern.")
        fields = args[:-1]
        pattern = compile_regex(args[-1])

        def predicate(record: RecordView) -> bool:
            for field in fields:
                for value in record.get_values(field):
                    if pattern.match(value or ""):
                        return True
            return False

        return predicate

    def _build_includes(self, args: list[str]) -> Callable[[RecordView], bool]:
        if len(args) < 2:
            raise QueryError("includes requires at least one field and a pattern.")
        fields = args[:-1]
        pattern = compile_regex(args[-1])

        def predicate(record: RecordView) -> bool:
            for field in fields:
                for value in record.get_values(field):
                    if pattern.search(value or ""):
                        return True
            return False

        return predicate

    def _build_equals(self, args: list[str]) -> Callable[[RecordView], bool]:
        if len(args) < 2:
            raise QueryError("equals requires a field and a value.")
        field = args[0]
        target = args[1]

        def predicate(record: RecordView) -> bool:
            return any(value == target for value in record.get_values(field))

        return predicate

    def _build_more(self, args: list[str]) -> Callable[[RecordView], bool]:
        if len(args) < 2:
            raise QueryError("more requires a field and a value.")
        field = args[0]
        target = args[1]

        def predicate(record: RecordView) -> bool:
            for value in record.get_values(field):
                if compare_values(value, target, ">="):
                    return True
            return False

        return predicate

    def _build_less(self, args: list[str]) -> Callable[[RecordView], bool]:
        if len(args) < 2:
            raise QueryError("less requires a field and a value.")
        field = args[0]
        target = args[1]

        def predicate(record: RecordView) -> bool:
            for value in record.get_values(field):
                if compare_values(value, target, "<="):
                    return True
            return False

        return predicate

    def _build_exists(self, args: list[str]) -> Callable[[RecordView], bool]:
        if len(args) != 1:
            raise QueryError("exists requires exactly one field.")
        field = args[0]

        def predicate(record: RecordView) -> bool:
            return any((value or "").strip() for value in record.get_values(field))

        return predicate

    def _build_any(self, args: list[str]) -> Callable[[RecordView], bool]:
        if len(args) < 2:
            raise QueryError("any requires a field followed by values.")
        field = args[0]
        values = args[1:]

        def predicate(record: RecordView) -> bool:
            field_values = set(record.get_list_values(field))
            return any(value in field_values for value in values)

        return predicate

    def _build_all(self, args: list[str]) -> Callable[[RecordView], bool]:
        if len(args) < 2:
            raise QueryError("all requires a field followed by values.")
        field = args[0]
        values = args[1:]

        def predicate(record: RecordView) -> bool:
            field_values = set(record.get_list_values(field))
            return all(value in field_values for value in values)

        return predicate

    def _build_one(self, args: list[str]) -> Callable[[RecordView], bool]:
        if len(args) < 2:
            raise QueryError("one requires a field followed by values.")
        field = args[0]
        values = args[1:]

        def predicate(record: RecordView) -> bool:
            field_values = record.get_values(field)
            return any(value in values for value in field_values)

        return predicate

    def _build_dt(self, args: list[str]) -> Callable[[RecordView], bool]:
        if len(args) < 2:
            raise QueryError("dt requires a field and expression.")
        field = args[0]
        operator = args[1]

        if operator == "?" and len(args) >= 3:
            kind = args[2].lower()
            if kind not in ("date", "time"):
                raise QueryError("dt ? expects 'date' or 'time'.")

            def predicate(record: RecordView) -> bool:
                values = record.get_datetime_values(field)
                if kind == "date":
                    return any(is_date_only for _, is_date_only in values)
                return any(not is_date_only for _, is_date_only in values)

            return predicate

        if operator not in (">", "<", "="):
            raise QueryError("dt comparison must use '>', '<', or '='.")
        if len(args) < 3:
            raise QueryError("dt comparison missing value.")
        target_value = parse_query_datetime(args[2])

        def predicate(record: RecordView) -> bool:
            values = record.get_datetime_values(field)
            for value, is_date_only in values:
                record_value = value.date() if is_date_only and isinstance(value, datetime) else value
                target = (
                    target_value.date()
                    if is_date_only and isinstance(target_value, datetime)
                    else target_value
                )
                if compare_dates(record_value, target, operator):
                    return True
            return False

        return predicate


class QueryEngine:
    """Entry point to parse and evaluate queries."""

    def __init__(self) -> None:
        self.parser = QueryParser()

    def run(
        self,
        text: str,
        records: Iterable[dict],
    ) -> QueryResponse:
        plan, info_id = self.parser.parse(text)
        if info_id is not None:
            return QueryResponse(matches=[], info_id=info_id)

        matches: list[QueryMatch] = []
        for record in records:
            view = RecordView(
                record_id=record.get("id", 0),
                itemtype=record.get("itemtype", ""),
                subject=record.get("subject", ""),
                tokens=record.get("tokens", []),
            )
            if plan.matches(view):
                matches.append(
                    QueryMatch(
                        record_id=view.record_id,
                        itemtype=view.itemtype,
                        subject=view.subject or "(untitled)",
                    )
                )
        return QueryResponse(matches=matches, info_id=None)


def compile_regex(pattern: str) -> re.Pattern[str]:
    try:
        return re.compile(pattern, flags=re.IGNORECASE)
    except re.error as exc:
        raise QueryError(f"Invalid regular expression: {pattern}") from exc


def normalize_field(name: str) -> str:
    name = (name or "").strip().lower()
    if name.startswith("@"):
        name = name[1:]
    if name in ("subject", "summary"):
        return "subject"
    if name in ("itemtype", "type", "item"):
        return "itemtype"
    return name


def negate_predicate(predicate: Callable[[RecordView], bool]) -> Callable[[RecordView], bool]:
    def _wrapped(record: RecordView) -> bool:
        return not predicate(record)

    return _wrapped


def compare_values(left: str, right: str, operator: str) -> bool:
    left_num = to_number(left)
    right_num = to_number(right)

    if left_num is not None and right_num is not None:
        lhs = left_num
        rhs = right_num
    else:
        lhs = left or ""
        rhs = right or ""

    if operator == ">=":
        return lhs >= rhs
    if operator == "<=":
        return lhs <= rhs
    raise QueryError(f"Unsupported comparison operator: {operator}")


def to_number(value: str) -> float | None:
    try:
        return float(value)
    except (TypeError, ValueError):
        return None


def parse_field_datetime(value: str) -> tuple[date | datetime, bool] | None:
    raw = (value or "").strip()
    if not raw:
        return None
    try:
        parsed = dt_parser.parse(raw)
    except Exception:
        return None
    is_date_only = bool(FIELD_REGEX_DATE.fullmatch(raw))
    if is_date_only:
        return (parsed.date(), True)
    return (parsed, False)


def parse_query_datetime(text: str) -> date | datetime:
    parts = [p.strip() for p in text.split("-") if p.strip()]
    if len(parts) < 3:
        raise QueryError("Datetime comparisons require yyyy-mm-dd or yyyy-mm-dd-HH-MM.")
    try:
        year = int(parts[0])
        month = int(parts[1])
        day = int(parts[2])
    except ValueError as exc:
        raise QueryError("Datetime components must be numeric.") from exc

    if len(parts) == 3:
        return date(year, month, day)

    try:
        hour = int(parts[3])
        minute = int(parts[4]) if len(parts) > 4 else 0
    except ValueError as exc:
        raise QueryError("Time components must be numeric.") from exc
    return datetime(year, month, day, hour, minute)


def compare_dates(
    record_value: date | datetime,
    target_value: date | datetime,
    operator: str,
) -> bool:
    if isinstance(record_value, datetime) and isinstance(target_value, date):
        target_value = datetime(
            target_value.year,
            target_value.month,
            target_value.day,
            0,
            0,
        )
    if isinstance(record_value, date) and isinstance(target_value, datetime):
        record_value = datetime(
            record_value.year,
            record_value.month,
            record_value.day,
            0,
            0,
        )

    if operator == ">":
        return record_value > target_value
    if operator == "<":
        return record_value < target_value
    if operator == "=":
        return record_value == target_value
    raise QueryError(f"Unsupported datetime operator: {operator}")



def test_urgency_rebuilds_when_priority_config_changes(test_controller):
    db = test_controller.db_manager

    # Seed derived-state entries.
    db.populate_dependent_tables(force=True)
    before = db._get_state_value("urgency", {}) or {}
    before_cfg = before.get("config_version")
    assert before_cfg

    # Change urgency config only (no record changes).
    priority_cfg = test_controller.env.config.urgency.priority.root
    priority_cfg["first"] = float(priority_cfg.get("first", 10.0)) + 1.0

    db.populate_dependent_tables(force=False)
    after = db._get_state_value("urgency", {}) or {}
    after_cfg = after.get("config_version")

    assert after_cfg
    assert after_cfg != before_cfg



import click
from rich.table import Table
from rich.console import Console

# from rich.color import Color
from rich.style import Style
from colorsys import rgb_to_hls

import subprocess

import io

try:
    from .named_colors import css_named_colors
except Exception:  # pragma: no cover - fallback for running from src/tklr
    from named_colors import css_named_colors


def sort_colors(
    color_dict: dict[str, str], method: str = "name"
) -> list[tuple[str, str]]:
    if method == "hue":

        def hex_to_hue(hex_value: str):
            hex_value = hex_value.lstrip("#")
            r, g, b = tuple(int(hex_value[i : i + 2], 16) / 255.0 for i in (0, 2, 4))
            h, _, _ = rgb_to_hls(r, g, b)
            return h

        return sorted(color_dict.items(), key=lambda item: hex_to_hue(item[1]))
    return sorted(color_dict.items())  # default name


def get_contrasting_text_color(hex_color: str) -> str:
    hex_color = hex_color.lstrip("#")
    r, g, b = [int(hex_color[i : i + 2], 16) for i in (0, 2, 4)]
    brightness = (r * 299 + g * 587 + b * 114) / 1000
    return "black" if brightness > 128 else "white"



@click.command()
@click.option(
    "--sort",
    type=click.Choice(["name", "hue"], case_sensitive=False),
    default="name",
    help="Sort colors name or by hue",
)
def show_colors(sort):
    with subprocess.Popen(["less", "-R"], stdin=subprocess.PIPE) as proc:
        # Wrap binary stdin as a text stream
        with io.TextIOWrapper(proc.stdin, encoding="utf-8") as text_stream:
            console = Console(
                file=text_stream, force_terminal=True, color_system="truecolor"
            )

            table = Table(title=f"CSS Named Colors (--sort {sort})")
            table.add_column("Color Name")
            table.add_column("Hex Value")

            for name, hex_val in sort_colors(css_named_colors, sort):
                name_style = Style(color=hex_val)
                hex_style = Style(
                    color=get_contrasting_text_color(hex_val), bgcolor=hex_val
                )
                table.add_row(f"[{hex_val}]{name}[/]", f"[{hex_style}]{hex_val}[/]")

            console.print(table)


if __name__ == "__main__":
    show_colors()



from __future__ import annotations
from textual.widgets import Tree
from rich.text import Text

"""
- Actions:
    - [bold]?[/bold]: Display this help screen
    - [bold]/[/bold]: Search
    - [bold]^Q[/bold]: Quit
    - [bold]^S[/bold]: Screenshot
    - [bold]a,b,...,z[/bold]: Display details for reminder with corresponding tag
    - [bold]A,B,...[/bold]: Display corresponding view
- Views: 
    - [bold]A[/bold]: Agenda 
    - [bold]B[/bold]: Bins 
    - [bold]C[/bold]: Completions
    - [bold]F[/bold]: Find 
    - [bold]H[/bold]: Hash Tags
    - [bold]L[/bold]: Last Instances
    - [bold]N[/bold]: Next Instances
    - [bold]R[/bold]: Remaining Alerts for Today
    - [bold]W[/bold]: Weeks
"""

HELP_SCHEMA = {
    "Actions": {
        "bindings": {
            "?": "Display this help screen",
            "/": "Search",
            "^Q": "Quit",
            "^S": "Screenshot",
            "a,b,...,z": "Display details for reminder with the corresponding tag",
            "A,B,...": "Display corresponding view",
        },
    },
    "Views": {
        "A: Agenda": {
            "description": "Shows the next 3 days of events + urgency-ranked tasks.",
            "bindings": {
                "→": "Next page",
                "←": "Previous page",
                "↑": "scroll up",
                "↓": "scroll down",
            },
        },
        "B: Bins": {},
        "C: Completions": {},
        "F: Find": {},
        "H: Hash Tags": {},
        "L: Last Instances": {},
        "N: Next Instances": {},
        "R: Remaining Alerts for Today": {},
        "W: Weeks": {},
    },
}
{}

# Hierarchical help schema for Tklr UI and CLI.
HELP_SCHEMA = {
    "UI (Textual)": {
        "Agenda View": {
            "description": "Shows 3-day pages + urgency-ranked tasks.",
            "bindings": {
                ">": "Next page",
                "<": "Previous page",
                "t": "Toggle tasks pane",
            },
        },
        "4-Week Table": {
            "description": "Calendar grid of 28 days with busy bars.",
            "bindings": {
                "j": "Jump to today",
                "w": "Cycle week focus",
            },
        },
        "Screens": {
            "BinHierarchyScreen": {
                "description": "Expandable bin navigation screen.",
                "bindings": {
                    "escape": "Dismiss",
                    "e": "Expand to depth",
                    "c": "Collapse node",
                },
            },
        },
    },
    "CLI (Click)": {
        "Bidding Simulation": {
            "description": "Bridge 2/1 Game Forcing with Stayman + Jacoby Transfers.",
            "bindings": {
                "b": "Begin bidding",
                "c": "Continue bidding",
                "n/e/s/w": "Reveal hand",
            },
        },
        "@-tokens": {
            "description": "Symbolic tokens for reminders, dates, bidding, urgency.",
            "items": {
                "@r": "New reminder record",
                "@f": "First instance datetime",
            },
        },
    },
    "Config & Environment": {
        "config.toml": {
            "description": "User preferences + conventions + recurrence rules.",
        },
        "tklr.db": {
            "description": "SQLite persistence for records, datetimes, bins, urgency.",
        },
    },
    "Some Feature": {
        "paragraph": (
            "This is one or more sentences of help text. It is not wrapped here. "
            "Rich + Textual will wrap it later based on terminal width."
        ),
        "Subfeature A": {
            "paragraph": "Even nested paragraphs are fine. They are attached under their node.",
        },
    },
}


def build_help_tree() -> Tree:
    """Build a Textual Tree from HELP_SCHEMA with paragraph leafs."""
    tree = Tree("Help", id="help-tree")

    def walk(parent, key: str, node_def: object) -> None:
        # Branch node (dict)
        if isinstance(node_def, dict):
            node = parent.add(key)

            # 1) Attach paragraph guidance if present on this branch
            if "paragraph" in node_def:
                node.add(Text(node_def["paragraph"], style="dim"))

            # 2) Recurse for real children (excluding structural keys)
            for k, v in node_def.items():
                if k in ("paragraph", "description", "bindings", "items"):
                    continue
                walk(node, k, v)

            # 3) Attach structured binding help under a collapsible group
            if "bindings" in node_def and isinstance(node_def["bindings"], dict):
                bind_group = node.add("⌨ Bindings")
                for bkey, bdesc in node_def["bindings"].items():
                    # bindings arrows kept short, children wrap if long
                    bind_group.add(Text(f"{bkey} → {bdesc}", style="bold"))

            # 4) Attach item lists (like token glossaries) under their own group
            if "items" in node_def and isinstance(node_def["items"], dict):
                item_group = node.add("🔖 Items")
                for ikey, idesc in node_def["items"].items():
                    item_group.add(Text(f"{ikey} → {idesc}", style="dim"))

        # Leaf paragraph (string)
        else:
            node = parent.add(key)
            # Attach inner string or treat paragraph payload as text
            if isinstance(node_def, str):
                node.add(Text(node_def, style="dim"))

    # Populate tree from schema
    for topic, definition in HELP_SCHEMA.items():
        walk(tree.root, topic, definition)

    # Default UX: expand only the root level
    if tree.root is not None:
        tree.root.expand()

    return tree


if __name__ == "__main__":
    tree = build_help_tree()
    tree.dump()



import importlib

import pytest
from click.testing import CliRunner

from tklr.model import DatabaseManager
from tklr.tklr_env import TklrConfig, TklrEnvironment, save_config_from_template


def _prepare_home(monkeypatch, tmp_path, cli_rich: bool):
    home = tmp_path / "home"
    home.mkdir()

    monkeypatch.setenv("TKLR_HOME", str(home))
    monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)

    env = TklrEnvironment()
    config = TklrConfig()
    config.ui.cli_rich = cli_rich
    save_config_from_template(config, env.config_path)
    dbm = DatabaseManager(str(env.db_path), env)
    dbm.conn.close()
    return env


def _reload_cli():
    import tklr.cli.main as cli_main

    return importlib.reload(cli_main)


class FakeConsole:
    instances = []

    def __init__(self, *args, **kwargs):
        self.kwargs = kwargs
        type(self).instances.append(self)

    def print(self, *args, **kwargs):
        return None


class _FakeDBManager:
    def get_events_for_period(self, start_dt, end_dt):
        return []


class FakeController:
    def __init__(self, db_path, env):
        self.db_manager = _FakeDBManager()
        self.AMPM = env.config.ui.ampm

    def get_agenda(self, yield_rows=False):
        return []


@pytest.mark.unit
@pytest.mark.parametrize("command", ["agenda", "days", "weeks"])
def test_cli_rich_uses_config_default(monkeypatch, tmp_path, command):
    _prepare_home(monkeypatch, tmp_path, cli_rich=True)
    cli_main = _reload_cli()
    FakeConsole.instances = []
    monkeypatch.setattr(cli_main, "Console", FakeConsole)
    monkeypatch.setattr(cli_main, "Controller", FakeController)

    runner = CliRunner()
    result = runner.invoke(cli_main.cli, [command])

    assert result.exit_code == 0
    assert len(FakeConsole.instances) == 1
    assert FakeConsole.instances[0].kwargs["markup"] is True
    assert FakeConsole.instances[0].kwargs["no_color"] is False


@pytest.mark.unit
@pytest.mark.parametrize("command", ["agenda", "days", "weeks"])
def test_cli_plain_overrides_config_default(monkeypatch, tmp_path, command):
    _prepare_home(monkeypatch, tmp_path, cli_rich=True)
    cli_main = _reload_cli()
    FakeConsole.instances = []
    monkeypatch.setattr(cli_main, "Console", FakeConsole)
    monkeypatch.setattr(cli_main, "Controller", FakeController)

    runner = CliRunner()
    result = runner.invoke(cli_main.cli, [command, "--plain"])

    assert result.exit_code == 0
    assert len(FakeConsole.instances) == 1
    assert FakeConsole.instances[0].kwargs["markup"] is False
    assert FakeConsole.instances[0].kwargs["no_color"] is True


@pytest.mark.unit
@pytest.mark.parametrize("command", ["agenda", "days", "weeks"])
def test_cli_rich_overrides_plain_config_default(monkeypatch, tmp_path, command):
    _prepare_home(monkeypatch, tmp_path, cli_rich=False)
    cli_main = _reload_cli()
    FakeConsole.instances = []
    monkeypatch.setattr(cli_main, "Console", FakeConsole)
    monkeypatch.setattr(cli_main, "Controller", FakeController)

    runner = CliRunner()
    result = runner.invoke(cli_main.cli, [command, "--rich"])

    assert result.exit_code == 0
    assert len(FakeConsole.instances) == 1
    assert FakeConsole.instances[0].kwargs["markup"] is True
    assert FakeConsole.instances[0].kwargs["no_color"] is False



import time
from datetime import datetime

import pytest

from tklr.tklr_env import TklrEnvironment
from tklr.model import DatabaseManager
from tklr.item import Item


@pytest.fixture
def isolated_env(tmp_path, monkeypatch):
    """Create a throwaway TKLR_HOME so tests never touch local data."""
    home = tmp_path / "tklr-home"
    monkeypatch.delenv("TKLR_HOME", raising=False)
    monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
    monkeypatch.setenv("TKLR_HOME", str(home))

    env = TklrEnvironment()
    env.ensure(init_config=True, init_db_fn=None)
    return env


@pytest.mark.parametrize("tz_name", ["Etc/GMT+5"])
def test_rdate_rows_use_local_time(isolated_env, monkeypatch, tz_name):
    """
    Regression test: RDATE-only reminders must emit local-naive timestamps
    into DateTimes even though the rruleset stores UTC ('...Z').
    """

    monkeypatch.setenv("TZ", tz_name)
    if hasattr(time, "tzset"):
        time.tzset()

    dbm = DatabaseManager(
        str(isolated_env.db_path),
        isolated_env,
        reset=True,
        auto_populate=False,
    )

    single = Item(
        env=isolated_env,
        raw="* single @s 2026-01-08 11:00 @e 1h",
        final=True,
    )
    recurring = Item(
        env=isolated_env,
        raw="* repeating @s 2026-01-08 11:00 @e 1h @r w",
        final=True,
    )

    single_id = dbm.add_item(single)
    repeat_id = dbm.add_item(recurring)

    dbm.generate_datetimes_for_record(single_id, clear_existing=True)
    dbm.generate_datetimes_for_record(repeat_id, clear_existing=True)
    dbm.conn.commit()

    (rruleset_single,) = dbm.cursor.execute(
        "SELECT rruleset FROM Records WHERE id=?", (single_id,)
    ).fetchone()
    assert "RDATE:20260108T1600Z" in rruleset_single

    rows = dbm.cursor.execute(
        "SELECT start_datetime FROM DateTimes WHERE record_id=? ORDER BY start_datetime",
        (single_id,),
    ).fetchall()
    assert rows == [("20260108T1100",)]

    repeat_rows = dbm.cursor.execute(
        "SELECT start_datetime FROM DateTimes WHERE record_id=? ORDER BY start_datetime LIMIT 2",
        (repeat_id,),
    ).fetchall()
    assert repeat_rows[0][0].endswith("1100")
    assert repeat_rows[1][0].endswith("1100")

    dbm.conn.close()


@pytest.mark.parametrize("tz_name", ["US/Eastern"])
def test_aware_weekly_recurrence_keeps_local_wall_time_across_dst(
    isolated_env, monkeypatch, tz_name
):
    """
    Aware recurring reminders are stored in UTC, but each generated instance
    should be expanded in the record's timezone before conversion back to the
    viewer's local timezone. This keeps an 8:00 AM US/Eastern series at 8:00 AM
    after the March 8, 2026 DST transition.
    """

    monkeypatch.setenv("TZ", tz_name)
    if hasattr(time, "tzset"):
        time.tzset()

    dbm = DatabaseManager(
        str(isolated_env.db_path),
        isolated_env,
        reset=True,
        auto_populate=False,
    )

    item = Item(
        env=isolated_env,
        raw=(
            "* pickleball @s 2026-01-05 8:00AM @e 90m "
            "@r w &w MO, TH @- 20260129T0800"
        ),
        final=True,
    )
    record_id = dbm.add_item(item)

    dbm.generate_datetimes_for_record(
        record_id,
        clear_existing=True,
        window=(datetime(2026, 3, 1), datetime(2026, 3, 31, 23, 59)),
    )
    dbm.conn.commit()

    (rruleset,) = dbm.cursor.execute(
        "SELECT rruleset FROM Records WHERE id=?",
        (record_id,),
    ).fetchone()
    assert "DTSTART:20260105T1300Z" in rruleset
    assert "EXDATE:20260129T1300Z" in rruleset

    rows = dbm.cursor.execute(
        "SELECT start_datetime FROM DateTimes WHERE record_id=? ORDER BY start_datetime",
        (record_id,),
    ).fetchall()
    assert rows == [
        ("20260302T0800",),
        ("20260305T0800",),
        ("20260309T0800",),
        ("20260312T0800",),
        ("20260316T0800",),
        ("20260319T0800",),
        ("20260323T0800",),
        ("20260326T0800",),
        ("20260330T0800",),
    ]

    dbm.conn.close()


def test_repetitions_fallback_handles_aware_rrules_without_datetime_rows(
    test_controller, item_factory, freeze_at
):
    item = item_factory(
        "* test for mel @s 2026-02-27 8:00AM @r w @a 0m: n",
        final=True,
    )
    record_id = test_controller.add_item(item)

    test_controller.db_manager.cursor.execute(
        "DELETE FROM DateTimes WHERE record_id = ?",
        (record_id,),
    )
    test_controller.db_manager.conn.commit()

    with freeze_at("2026-03-13 11:37:00"):
        title, lines = test_controller.get_record_repetitions(record_id, limit=3)

    assert title == "Repetitions for test for mel"
    assert lines
    assert not any("Unable to parse rruleset" in line for line in lines)
    assert lines[0].startswith("Next ")


def test_repetitions_can_start_from_selected_instance_with_datetime_rows(
    test_controller, item_factory, freeze_at
):
    item = item_factory("* weekly check-in @s 2026-02-27 8:00AM @r w", final=True)
    record_id = test_controller.add_item(item)

    test_controller.db_manager.generate_datetimes_for_record(
        record_id,
        clear_existing=True,
        window=(datetime(2026, 2, 1), datetime(2026, 4, 30, 23, 59)),
    )
    test_controller.db_manager.conn.commit()

    with freeze_at("2026-03-13 11:37:00"):
        title, lines = test_controller.get_record_repetitions(
            record_id,
            limit=3,
            instance_ts="20260327T0800",
        )

    assert title == "Repetitions for weekly check-in"
    assert lines[0] == "Next 3 occurrence(s):"
    assert lines[1] == f"  • {test_controller.fmt_user(datetime(2026, 3, 27, 8, 0))}"
    assert lines[2] == f"  • {test_controller.fmt_user(datetime(2026, 4, 3, 8, 0))}"


def test_repetitions_fallback_can_start_from_selected_instance(
    test_controller, item_factory, freeze_at
):
    item = item_factory("* weekly fallback @s 2026-02-27 8:00AM @r w", final=True)
    record_id = test_controller.add_item(item)

    test_controller.db_manager.cursor.execute(
        "DELETE FROM DateTimes WHERE record_id = ?",
        (record_id,),
    )
    test_controller.db_manager.conn.commit()

    with freeze_at("2026-03-13 11:37:00"):
        title, lines = test_controller.get_record_repetitions(
            record_id,
            limit=3,
            instance_ts="20260327T0800",
        )

    assert title == "Repetitions for weekly fallback"
    assert lines[0] == "Next 3 occurrence(s):"
    assert lines[1] == f"  • {test_controller.fmt_user(datetime(2026, 3, 27, 8, 0))}"



from datetime import datetime, timedelta, timezone

from tklr.item import Item


def _dt(y: int, m: int, d: int, hh: int = 0, mm: int = 0) -> datetime:
    return datetime(y, m, d, hh, mm, tzinfo=timezone.utc)


def _record_id_for_entry(test_controller, test_env, entry: str) -> int:
    item = Item(raw=entry, env=test_env, final=True, controller=test_controller)
    assert item.parse_ok, item.parse_message
    return test_controller.add_item(item)


def _tag_count(pages):
    return sum(len(tag_map) for _rows, tag_map in pages)


def test_completion_retention_prunes_infinite_repeating_tasks(test_controller, test_env):
    test_controller.env.config.num_completions = 2
    rid = _record_id_for_entry(test_controller, test_env, "~ repeat forever @s 2025-01-01 @r d")

    db = test_controller.db_manager
    db.add_completion(rid, (_dt(2025, 1, 1, 10), None))
    db.add_completion(rid, (_dt(2025, 1, 2, 10), None))
    db.add_completion(rid, (_dt(2025, 1, 3, 10), None))

    db.cursor.execute("SELECT COUNT(*) FROM Completions WHERE record_id = ?", (rid,))
    assert db.cursor.fetchone()[0] == 2


def test_completion_retention_does_not_prune_finite_repeating_tasks(
    test_controller, test_env
):
    test_controller.env.config.num_completions = 2
    rid = _record_id_for_entry(
        test_controller,
        test_env,
        "~ repeat three times @s 2025-01-01 @r d &c 3",
    )

    db = test_controller.db_manager
    db.add_completion(rid, (_dt(2025, 1, 1, 10), None))
    db.add_completion(rid, (_dt(2025, 1, 2, 10), None))
    db.add_completion(rid, (_dt(2025, 1, 3, 10), None))

    db.cursor.execute("SELECT COUNT(*) FROM Completions WHERE record_id = ?", (rid,))
    assert db.cursor.fetchone()[0] == 3


def test_completion_retention_zero_keeps_all(test_controller, test_env):
    test_controller.env.config.num_completions = 0
    rid = _record_id_for_entry(test_controller, test_env, "~ keep all @s 2025-01-01 @r d")

    db = test_controller.db_manager
    db.add_completion(rid, (_dt(2025, 1, 1, 10), None))
    db.add_completion(rid, (_dt(2025, 1, 2, 10), None))
    db.add_completion(rid, (_dt(2025, 1, 3, 10), None))

    db.cursor.execute("SELECT COUNT(*) FROM Completions WHERE record_id = ?", (rid,))
    assert db.cursor.fetchone()[0] == 3


def test_delete_completion_by_id(test_controller, test_env):
    rid = _record_id_for_entry(test_controller, test_env, "~ delete one @s 2025-01-01")
    db = test_controller.db_manager
    db.add_completion(rid, (_dt(2025, 1, 1, 10), None))
    db.add_completion(rid, (_dt(2025, 1, 2, 10), None))

    rows = db.get_all_completions_with_ids()
    assert len(rows) == 2
    completion_id = rows[0][0]

    assert db.delete_completion(completion_id) is True
    assert db.delete_completion(completion_id) is False

    remaining_ids = [row[0] for row in db.get_all_completions_with_ids()]
    assert completion_id not in remaining_ids
    assert len(remaining_ids) == 1


def test_completions_view_tag_map_includes_completion_id(test_controller, test_env):
    rid = _record_id_for_entry(
        test_controller, test_env, "~ completion tags @s 2025-01-01"
    )
    db = test_controller.db_manager
    db.add_completion(rid, (_dt(2025, 1, 3, 10), None))

    pages, _header = test_controller.get_completions()
    assert pages
    _rows, tag_map = pages[0]
    assert tag_map

    first_payload = next(iter(tag_map.values()))
    assert first_payload[0] == rid
    completion_id = first_payload[2]
    assert isinstance(completion_id, int)

    db.cursor.execute("SELECT id FROM Completions WHERE record_id = ?", (rid,))
    known_ids = {row[0] for row in db.cursor.fetchall()}
    assert completion_id in known_ids


def test_completions_view_omits_midnight_time(test_controller, test_env):
    rid = _record_id_for_entry(
        test_controller, test_env, "~ completion formatting @s 2025-01-01"
    )
    test_controller.dayfirst = True
    test_controller.yearfirst = False
    test_controller.two_digit_year = False
    test_controller.AMPM = True

    completed_local = datetime(2025, 1, 5, 10, 15).astimezone()
    test_controller.db_manager.add_completion(rid, (completed_local, None))

    pages, _header = test_controller.get_completions()
    assert pages
    rows, _tag_map = pages[0]
    record_rows = [row for row in rows if "[/not bold]   [" in row]
    assert record_rows
    row_text = record_rows[0]
    assert "[not bold]05-01-2025[/not bold]" in row_text, row_text
    assert "00:00" not in row_text
    assert "12a" not in row_text


def test_completions_view_keeps_non_midnight_time(test_controller, test_env):
    rid = _record_id_for_entry(test_controller, test_env, "~ completion time @s 2025-01-01")
    test_controller.dayfirst = False
    test_controller.yearfirst = True
    test_controller.two_digit_year = True
    test_controller.AMPM = False

    test_controller.db_manager.add_completion(rid, (_dt(2025, 1, 5, 0, 0), None))
    cursor = test_controller.db_manager.cursor
    cursor.execute(
        "UPDATE Completions SET completed = ? WHERE record_id = ?",
        ("2025-01-05 10:15", rid),
    )
    test_controller.db_manager.conn.commit()

    pages, _header = test_controller.get_completions()
    assert pages
    rows, _tag_map = pages[0]
    record_rows = [row for row in rows if "[/not bold]   [" in row]
    assert record_rows
    assert "[not bold]25-01-05 10:15[/not bold]" in record_rows[0], record_rows[0]


def test_completions_view_paginates_when_tags_exhausted(test_controller, test_env):
    base_local = datetime(2025, 1, 1, 9, 0).astimezone()
    for i in range(30):
        rid = _record_id_for_entry(test_controller, test_env, f"~ completion row {i}")
        completed_local = base_local + timedelta(minutes=i)
        test_controller.db_manager.add_completion(rid, (completed_local, None))

    pages, _header = test_controller.get_completions()
    assert len(pages) >= 2
    assert _tag_count(pages) == 30


def test_completions_view_dims_unchanged_year_month_components(
    test_controller, test_env
):
    test_controller.dayfirst = False
    test_controller.yearfirst = True
    test_controller.two_digit_year = False
    test_controller.AMPM = False

    rid_new = _record_id_for_entry(test_controller, test_env, "~ newer completion")
    rid_old = _record_id_for_entry(test_controller, test_env, "~ older completion")
    test_controller.db_manager.add_completion(rid_new, (_dt(2025, 1, 5, 10, 15), None))
    test_controller.db_manager.add_completion(rid_old, (_dt(2025, 1, 3, 10, 15), None))

    pages, _header = test_controller.get_completions()
    assert pages
    rows, _tag_map = pages[0]
    record_rows = [row for row in rows if "[/not bold]   [" in row]
    assert len(record_rows) >= 2

    second_row = record_rows[1]
    dim = test_controller.dim_style
    assert f"[{dim}]2025[/{dim}]" in second_row, second_row
    assert f"[{dim}]01[/{dim}]" in second_row, second_row



import pytest


@pytest.mark.integration
def test_record_urgency_components_include_weight_breakdown(
    test_controller, item_factory, freeze_at
):
    with freeze_at("2025-01-10 12:00:00"):
        item = item_factory(
            "~ submit report @s 2025-01-10 14:00 @p 1 @e 30m @d write and send"
        )
        assert item.parse_ok, item.parse_message
        record_id = test_controller.add_item(item)
        test_controller.db_manager.populate_dependent_tables(force=True)

        title, lines = test_controller.get_record_urgency_components(record_id)
        body = "\n".join(lines)

        assert title.startswith("Urgency components for")
        assert "urgency:" in body
        assert "components:" in body
        assert "due" in body
        assert "pastdue" in body


@pytest.mark.integration
def test_record_urgency_components_reports_missing_row_for_hidden_task(
    test_controller, item_factory, freeze_at
):
    with freeze_at("2025-01-10 12:00:00"):
        item = item_factory("~ hidden task @s 2025-01-20 10:00 @n 1d")
        assert item.parse_ok, item.parse_message
        record_id = test_controller.add_item(item)
        test_controller.db_manager.populate_dependent_tables(force=True)

        _title, lines = test_controller.get_record_urgency_components(record_id)
        assert lines
        assert "No urgency entry is currently available" in lines[0]



"""
Tests for basic Item parsing functionality.

These tests verify that Item objects can correctly parse entry strings
and extract basic attributes like type, summary, priority, etc.
"""

import pytest
from datetime import datetime, date
from tklr.item import Item
from tests.conftest import bin_path_contains_prefix


@pytest.mark.unit
class TestBasicParsing:
    """Test basic parsing of different item types."""

    def test_simple_task(self, item_factory):
        """Test parsing a simple task without any attributes."""
        item = item_factory("~ simple task")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.itemtype == "~"
        assert "simple task" in item.subject

    def test_simple_event(self, item_factory):
        """Test parsing a simple event."""
        item = item_factory("* simple event")

        assert not item.parse_ok, (
            f"Parse failed for '{item.entry}': {item.parse_message}"
        )
        assert item.itemtype == "*"
        assert "simple event" in item.subject

    def test_simple_journal(self, item_factory):
        """Test parsing a simple journal entry."""
        item = item_factory("% simple journal")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.itemtype == "%"
        assert "simple journal" in item.subject

    def test_draft_item(self, item_factory):
        """Test parsing a draft item."""
        item = item_factory("? draft reminder - no checks")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.itemtype == "?"

    def test_goal_item(self, item_factory):
        """Test parsing a goal item."""
        item = item_factory("! fitness goal")

        assert not item.parse_ok, (
            f"Parse failed for '{item.entry}': {item.parse_message}"
        )
        assert item.itemtype == "!"


@pytest.mark.unit
class TestPriorities:
    """Test priority parsing."""

    def test_priority_1(self, item_factory):
        """Test parsing priority 1 (highest)."""
        item = item_factory("~ high priority task @p 1")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.priority == 1

    def test_priority_2(self, item_factory):
        """Test parsing priority 2."""
        item = item_factory("~ medium priority task @p 2")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.priority == 2

    def test_priority_3(self, item_factory):
        """Test parsing priority 3."""
        item = item_factory("~ low priority task @p 3")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.priority == 3

    def test_priority_4(self, item_factory):
        """Test parsing priority 4."""
        item = item_factory("~ priority four @p 4")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.priority == 4

    def test_priority_5(self, item_factory):
        """Test parsing priority 5 (lowest)."""
        item = item_factory("~ priority five @p 5")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.priority == 5

    def test_no_priority(self, item_factory):
        """Test item without explicit priority."""
        item = item_factory("~ no priority task")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        # Check default priority handling


@pytest.mark.unit
class TestDescriptions:
    """Test description and tag parsing."""

    def test_simple_description(self, item_factory):
        """Test parsing item with description."""
        item = item_factory("~ task with details @d This is a detailed description")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert "This is a detailed description" in item.description

    def test_description_with_tags(self, item_factory):
        """Test parsing description with hashtags."""
        item = item_factory("~ task @d Description with #red #white #blue tags")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert "Description with" in item.description
        assert "#red" in item.description or "red" in str(item.tags)

    def test_multiline_description(self, frozen_time, item_factory):
        """Test parsing formatted multiline description."""
        # Time frozen to 2025-01-01 12:00:00 via frozen_time fixture

        entry = """% long formatted description @s 2025-01-01
    @d Title
    1. This
       i. with part one
       ii. and this
    2. And finally this.
    """
        item = item_factory(entry)

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert "Title" in item.description
        assert "This" in item.description

    def test_description_allows_email_address(self, item_factory):
        item = item_factory("~ task @d Contact dnlgrhm@pm.me for follow-up")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert "dnlgrhm@pm.me" in item.description
        detail_tokens = [
            tok
            for tok in item.relative_tokens
            if tok.get("t") == "@" and tok.get("k") == "d"
        ]
        assert detail_tokens
        assert "dnlgrhm@pm.me" in detail_tokens[0]["token"]

    def test_non_space_at_is_literal_but_space_at_starts_token(self, item_factory):
        item = item_factory("~ task @d Contact dnlgrhm@pm.me @s 2025-01-01")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert "dnlgrhm@pm.me" in item.description

        at_tokens = [tok for tok in item.relative_tokens if tok.get("t") == "@"]
        keys = [tok.get("k") for tok in at_tokens]
        assert "d" in keys
        assert "s" in keys


@pytest.mark.unit
class TestProjectLiveEditing:
    def test_incomplete_project_job_line_does_not_crash(self, item_factory):
        entry = """^ Project test
  @~ step 1
  @~
  @~ step 3
"""

        item = item_factory(entry, final=False)

        assert item.itemtype == "^"
        assert item.subject == "Project test"
        assert any(tok.get("k") == "~" for tok in item.relative_tokens)

    def test_project_job_requires_r_label_on_final_parse(self, item_factory):
        item = item_factory("^ Project test @~ step 1", final=True)

        assert not item.parse_ok
        assert "Each @~ job requires an &r label" in (item.parse_message or "")

    def test_project_job_missing_r_label_survives_second_finalize(self, item_factory):
        item = item_factory("^ Project test @~ step 1", final=True)
        item.finalize_record()

        assert not item.parse_ok
        assert "Each @~ job requires an &r label" in (item.parse_message or "")

    def test_live_project_schedule_feedback_is_not_hidden_by_missing_job(
        self, item_factory
    ):
        item = item_factory("^ project @s s", final=False)

        assert not item.parse_ok
        assert "Required keys not yet provided" not in (item.parse_message or "")
        assert "Error parsing 's'" in (item.parse_message or "")

    def test_live_project_bare_at_still_shows_required_and_available_keys(
        self, item_factory
    ):
        item = item_factory("^ project @", final=False)

        assert not item.parse_ok
        assert "@ available @-keys:" in (item.parse_message or "")
        assert "required: ~" in (item.parse_message or "")
        assert "optional:" in (item.parse_message or "")

    def test_live_project_rrule_trailing_amp_is_treated_as_incomplete_modifier(
        self, item_factory
    ):
        item = item_factory("^ project @s sat @r m &\n@~ step one", final=False)

        assert not item.parse_ok
        assert "supported frequency" not in (item.parse_message or "")
        assert "repetition &-key: enter &-key" in (item.parse_message or "")


@pytest.mark.unit
class TestInvitees:
    def test_event_invitees_parse(self, item_factory):
        item = item_factory("* planning session @s 2025-01-10 10:00 @i Alice, Bob")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.invitees == ["Alice", "Bob"]

    def test_event_invitees_require_non_empty_names(self, item_factory):
        item = item_factory("* planning session @s 2025-01-10 10:00 @i Alice, , Bob")

        assert not item.parse_ok
        assert item.last_result
        assert "comma separated list of non-empty names" in (item.last_result[1] or "")

    def test_invitees_not_allowed_for_tasks(self, item_factory):
        item = item_factory("~ prepare agenda @i Alice")

        assert not item.parse_ok
        assert "The use of @i is not supported" in (item.parse_message or "")


@pytest.mark.unit
class TestBins:
    """Test bin/category parsing."""

    def test_single_bin(self, item_factory):
        """Test parsing item with a single bin."""
        item = item_factory("% item @b work/projects")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.bin_paths is not None
        assert bin_path_contains_prefix(item.bin_paths, ["work", "projects"])
        # assert bin_path_contains_prefix(item.bin_paths, ["projects"])
        assert bin_path_contains_prefix(item.bin_paths, ["work"])

    def test_multiple_bins(self, item_factory):
        """Test parsing item with multiple bins."""
        item = item_factory("~ task @b errands/contexts @b urgent/tags")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.bin_paths is not None
        assert bin_path_contains_prefix(item.bin_paths, ["errands", "contexts"])
        assert bin_path_contains_prefix(item.bin_paths, ["urgent", "tags"])
        # assert ["urgent", "tags"] in item.bin_paths

    def test_complex_bin_path(self, item_factory):
        """Test parsing item with complex bin hierarchy."""
        item = item_factory("% note @b Churchill/quotations/library")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert bin_path_contains_prefix(
            item.bin_paths, ["Churchill", "quotations", "library"]
        )
        assert not bin_path_contains_prefix(item.bin_paths, ["Churchill", "library"])
        assert not bin_path_contains_prefix(item.bin_paths, ["quotations"])
        assert bin_path_contains_prefix(item.bin_paths, ["Churchill", "quotations"])
        assert bin_path_contains_prefix(item.bin_paths, ["Churchill"])

    def test_bin_with_year_month(self, item_factory):
        """Test parsing bin with year/month pattern."""
        item = item_factory("% entry @b 2025:10/2025/journal")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert bin_path_contains_prefix(item.bin_paths, ["2025:10", "2025", "journal"])


@pytest.mark.unit
class TestLogEntries:
    """Tests for log (-) reminders."""

    def test_log_defaults_schedule(self, frozen_time, item_factory):
        """Log entries without @s should auto-append the current time."""
        entry = "- hydration log"
        item = item_factory(entry)

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.auto_log_timestamp is not None

        expected_now = datetime.now().astimezone()
        assert item.auto_log_timestamp == expected_now

        s_tokens = [
            tok
            for tok in item.relative_tokens
            if tok.get("t") == "@" and tok.get("k") == "s"
        ]
        assert len(s_tokens) == 1
        assert "@s" in s_tokens[0]["token"]
        assert item.dtstart_str is not None

    def test_log_respects_explicit_schedule(self, item_factory):
        """If @s is provided, logs should not override it."""
        entry = "- hydration log @s 2025-02-01 09:30"
        item = item_factory(entry)

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.auto_log_timestamp is None

        s_tokens = [
            tok
            for tok in item.relative_tokens
            if tok.get("t") == "@" and tok.get("k") == "s"
        ]
        assert len(s_tokens) == 1
        assert "2025-02-01" in s_tokens[0]["token"]
        assert "9:30" in s_tokens[0]["token"]


@pytest.mark.unit
class TestUseLookup:
    """Ensure log-specific @u values come from the use registry."""

    def test_existing_use_parses(self, item_factory, use_factory):
        use_factory("Client Alpha")
        item = item_factory("- call recap @u Client Alpha")

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.use == "Client Alpha"
        assert item.use_id is not None

    def test_unknown_use_suggests(self, item_factory, use_factory):
        use_factory("Jones, Robert")
        use_factory("Jordan, Alex")
        use_factory("John, Sam")
        use_factory("Jules, Max")

        item = item_factory("- phoned client @u J")

        assert not item.parse_ok
        assert item.last_result
        assert "Matching entries:" in (item.last_result[1] or "")
        assert "Jones, Robert" in (item.last_result[1] or "")
        assert "Jordan, Alex" in (item.last_result[1] or "")
        assert "..." in (item.last_result[1] or "")


@pytest.mark.unit
class TestLiveAttributeMatching:
    def test_single_use_match_sets_live_replacement(self, item_factory, use_factory):
        use_factory("exercise.walking")

        item = item_factory("- jog @u e", final=False)

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.live_replacement is not None
        assert item.live_replacement[2] == "@u exercise.walking"

    def test_single_context_match_sets_live_replacement(
        self, item_factory, test_controller
    ):
        existing = item_factory("~ existing task @c Errands")
        test_controller.add_item(existing)

        item = item_factory("~ new task @c e", final=False)

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.live_replacement is not None
        assert item.live_replacement[2] == "@c Errands"
        context_token = next(
            tok
            for tok in item.relative_tokens
            if tok.get("t") == "@" and tok.get("k") == "c"
        )
        assert "Errands" in context_token.get("_matches", [])

    def test_single_location_match_sets_live_replacement(
        self, item_factory, test_controller
    ):
        existing = item_factory("~ existing task @l Home Office")
        test_controller.add_item(existing)

        item = item_factory("~ new task @l h", final=False)

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.live_replacement is not None
        assert item.live_replacement[2] == "@l Home Office"
        location_token = next(
            tok
            for tok in item.relative_tokens
            if tok.get("t") == "@" and tok.get("k") == "l"
        )
        assert "Home Office" in location_token.get("_matches", [])

    def test_single_bin_match_sets_live_replacement(self, item_factory, test_controller):
        seed = item_factory("% filing note @b Churchill/quotations/library")
        test_controller.add_item(seed)

        item = item_factory("~ new task @b chur", final=False)

        assert item.parse_ok, f"Parse failed for '{item.entry}': {item.parse_message}"
        assert item.live_replacement is not None
        assert item.live_replacement[2].startswith("@b Churchill/quotations/library")
        bin_token = next(
            tok
            for tok in item.relative_tokens
            if tok.get("t") == "@" and tok.get("k") == "b"
        )
        assert any(
            match.startswith("Churchill/quotations/library")
            for match in bin_token.get("_matches", [])
        )



from freezegun import freeze_time

from tklr.shared import bug_msg, log_msg
from tklr.tklr_env import TklrEnvironment, save_config_from_template


def _init_home_with_log_limit(tmp_path, monkeypatch, keep: int = 2):
    home = tmp_path / "tklr-home"
    monkeypatch.setenv("TKLR_HOME", str(home))
    env = TklrEnvironment()
    env.ensure(init_config=True)
    cfg = env.load_config()
    cfg.num_logs = keep
    save_config_from_template(cfg, env.config_path)
    return home


def test_daily_log_file_retention_for_log_and_bug(tmp_path, monkeypatch):
    home = _init_home_with_log_limit(tmp_path, monkeypatch, keep=2)

    for day in ("2026-01-01 10:00:00", "2026-01-02 10:00:00", "2026-01-03 10:00:00"):
        with freeze_time(day):
            log_msg("log entry")
            bug_msg("bug entry")

    log_files = sorted((home / "logs").glob("log_*.md"))
    bug_files = sorted((home / "logs").glob("bug_*.md"))

    assert [p.name for p in log_files] == ["log_260102.md", "log_260103.md"]
    assert [p.name for p in bug_files] == ["bug_260102.md", "bug_260103.md"]



from datetime import datetime
import re

import pytest


def _strip_markup(text: str) -> str:
    return re.sub(r"\[[^\]]+\]", "", text)


def _add_entry(test_controller, item_factory, entry: str) -> None:
    item = item_factory(entry)
    assert item.parse_ok, item.parse_message
    test_controller.add_item(item)


@pytest.mark.integration
def test_agenda_tasks_due_prefix_for_overdue_and_next_week(test_controller, item_factory):
    _add_entry(test_controller, item_factory, "~ overdue task @s 2025-01-06")
    _add_entry(test_controller, item_factory, "~ due today task @s 2025-01-10")
    _add_entry(test_controller, item_factory, "~ due soon task @s 2025-01-13")
    _add_entry(test_controller, item_factory, "~ due later task @s 2025-01-20")
    test_controller.db_manager.populate_dependent_tables(force=True)

    rows = test_controller.get_agenda_tasks(now=datetime(2025, 1, 10, 12, 0))
    texts = [_strip_markup(r["text"]) for r in rows if r.get("record_id") is not None]

    overdue = next(t for t in texts if "overdue task" in t)
    today = next(t for t in texts if "due today task" in t)
    soon = next(t for t in texts if "due soon task" in t)
    later = next(t for t in texts if "due later task" in t)

    assert "-4d overdue task" in overdue
    assert "+0d due today task" in today
    assert "+3d due soon task" in soon
    assert "due later task" in later
    assert "+10d" not in later


@pytest.mark.integration
def test_agenda_tasks_due_prefix_uses_current_date_not_time(test_controller, item_factory):
    _add_entry(test_controller, item_factory, "~ morning deadline @s 2025-01-10 09:00")
    test_controller.db_manager.populate_dependent_tables(force=True)

    rows = test_controller.get_agenda_tasks(now=datetime(2025, 1, 10, 18, 0))
    texts = [_strip_markup(r["text"]) for r in rows if r.get("record_id") is not None]
    row = next(t for t in texts if "morning deadline" in t)

    assert "+0d morning deadline" in row



from datetime import datetime, timedelta

import pytest


def _strip_markup(text: str) -> str:
    import re

    return re.sub(r"\[[^\]]+\]", "", text)


def _record_lines(pages):
    for page_rows, _ in pages:
        for row in page_rows:
            if "[/not bold]   [" in row:
                yield row


def _first_record_line(pages):
    for row in _record_lines(pages):
        return row
    raise AssertionError("No record rows found")


def _tag_count(pages):
    return sum(len(tag_map) for _rows, tag_map in pages)


@pytest.mark.unit
def test_last_view_shows_numeric_ymd_by_default(
    frozen_time, test_controller, item_factory
):
    test_controller.dayfirst = False
    test_controller.yearfirst = True
    test_controller.two_digit_year = True

    item = item_factory("* past event @s 2024-12-31 09:00 z none")
    assert item.parse_ok
    test_controller.add_item(item)
    test_controller.db_manager.populate_dependent_tables()

    pages, _ = test_controller.get_last()
    row_text = _first_record_line(pages)
    assert "24-12-31" in _strip_markup(row_text), row_text
    assert ":" not in row_text


@pytest.mark.unit
def test_next_view_shows_numeric_ymd_by_default(
    frozen_time, test_controller, item_factory
):
    test_controller.dayfirst = False
    test_controller.yearfirst = True
    test_controller.two_digit_year = True

    item = item_factory("* future event @s 2025-01-05 14:00 z none")
    assert item.parse_ok
    test_controller.add_item(item)
    test_controller.db_manager.populate_dependent_tables()

    pages, _ = test_controller.get_next()
    row_text = _first_record_line(pages)
    assert "25-01-05" in _strip_markup(row_text), row_text
    assert ":" not in row_text


@pytest.mark.unit
def test_last_view_honors_dayfirst_yearfirst_and_two_digit_year(
    frozen_time, test_controller, item_factory
):
    test_controller.dayfirst = True
    test_controller.yearfirst = False
    test_controller.two_digit_year = False

    item = item_factory("* past event @s 2024-12-31 09:00 z none")
    assert item.parse_ok
    test_controller.add_item(item)
    test_controller.db_manager.populate_dependent_tables()

    pages, _ = test_controller.get_last()
    row_text = _first_record_line(pages)
    assert "31-12-2024" in _strip_markup(row_text), row_text


@pytest.mark.unit
def test_next_view_honors_dayfirst_yearfirst_and_two_digit_year(
    frozen_time, test_controller, item_factory
):
    test_controller.dayfirst = True
    test_controller.yearfirst = False
    test_controller.two_digit_year = False

    item = item_factory("* future event @s 2025-01-05 14:00 z none")
    assert item.parse_ok
    test_controller.add_item(item)
    test_controller.db_manager.populate_dependent_tables()

    pages, _ = test_controller.get_next()
    row_text = _first_record_line(pages)
    assert "05-01-2025" in _strip_markup(row_text), row_text


@pytest.mark.unit
def test_next_view_stays_ascending(frozen_time, test_controller, item_factory):
    test_controller.dayfirst = False
    test_controller.yearfirst = True
    test_controller.two_digit_year = False

    for when in ["2025-01-05 09:00", "2025-01-03 09:00"]:
        item = item_factory(f"* future event @s {when} z none")
        assert item.parse_ok
        test_controller.add_item(item)
    test_controller.db_manager.populate_dependent_tables()

    pages, _ = test_controller.get_next()
    lines = list(_record_lines(pages))
    assert "2025-01-03" in _strip_markup(lines[0]), lines[0]
    assert "2025-01-05" in _strip_markup(lines[1]), lines[1]
    dim = test_controller.dim_style
    assert f"[{dim}]2025[/{dim}]" in lines[1], lines[1]
    assert f"[{dim}]01[/{dim}]" in lines[1], lines[1]
    assert f"[{dim}]-[/{dim}]" in lines[1], lines[1]


@pytest.mark.unit
def test_last_view_stays_descending(frozen_time, test_controller, item_factory):
    test_controller.dayfirst = False
    test_controller.yearfirst = True
    test_controller.two_digit_year = False

    for when in ["2024-12-30 09:00", "2024-12-31 09:00"]:
        item = item_factory(f"* past event @s {when} z none")
        assert item.parse_ok
        test_controller.add_item(item)
    test_controller.db_manager.populate_dependent_tables()

    pages, _ = test_controller.get_last()
    lines = list(_record_lines(pages))
    assert "2024-12-31" in _strip_markup(lines[0]), lines[0]
    assert "2024-12-30" in _strip_markup(lines[1]), lines[1]


@pytest.mark.unit
def test_next_view_paginates_when_tags_exhausted(
    frozen_time, test_controller, item_factory
):
    test_controller.dayfirst = False
    test_controller.yearfirst = True
    test_controller.two_digit_year = True

    base = datetime(2025, 1, 2, 9, 0)
    for i in range(30):
        when = (base + timedelta(days=i)).strftime("%Y-%m-%d %H:%M")
        item = item_factory(f"* future event {i} @s {when} z none")
        assert item.parse_ok
        test_controller.add_item(item)
    test_controller.db_manager.populate_dependent_tables()

    pages, _ = test_controller.get_next()
    assert len(pages) >= 2
    assert _tag_count(pages) == 30


@pytest.mark.unit
def test_last_view_paginates_when_tags_exhausted(
    frozen_time, test_controller, item_factory
):
    test_controller.dayfirst = False
    test_controller.yearfirst = True
    test_controller.two_digit_year = True

    base = datetime(2024, 12, 31, 9, 0)
    for i in range(30):
        when = (base - timedelta(days=i)).strftime("%Y-%m-%d %H:%M")
        item = item_factory(f"* past event {i} @s {when} z none")
        assert item.parse_ok
        test_controller.add_item(item)
    test_controller.db_manager.populate_dependent_tables()

    pages, _ = test_controller.get_last()
    assert len(pages) >= 2
    assert _tag_count(pages) == 30



from tklr.model import UrgencyComputer, td_str_to_seconds


def test_urgency_due_is_zero_when_due_is_beyond_due_interval(test_env):
    uc = UrgencyComputer(test_env)
    due_max = float(test_env.config.urgency.due.max)
    interval_seconds = td_str_to_seconds(test_env.config.urgency.due.interval)

    now = 1_000_000
    due_far = now + interval_seconds + 60
    due_score = uc.urgency_due(due_far, now)

    assert due_max > 0
    assert interval_seconds > 0
    assert due_score == 0.0


def test_urgency_due_reaches_max_at_due_datetime(test_env):
    uc = UrgencyComputer(test_env)
    due_max = float(test_env.config.urgency.due.max)
    now = 1_000_000

    due_score = uc.urgency_due(now, now)
    assert due_score == due_max
